@@ -1,7 +1,7 @@
 from __future__ import annotations
 
 import logging
-from asyncio import AbstractEventLoop, Event, Lock, Task
+from asyncio import AbstractEventLoop, CancelledError, Event, Lock, Task
 from datetime import datetime
 from enum import Enum
 
@@ -14,19 +14,27 @@ from bleak_retry_connector import (
     establish_connection,
 )
 from events import Events
-from transitions import Machine, State
+from transitions import EventData, Machine, State
 
+from pysnooz.advertisement import get_snooz_display_name
 from pysnooz.api import SnoozDeviceApi, SnoozDeviceState, UnknownSnoozState
 from pysnooz.commands import (
+    CommandProcessorState,
     SnoozCommandData,
     SnoozCommandProcessor,
     SnoozCommandResult,
     create_command_processor,
 )
+from pysnooz.const import UNEXPECTED_ERROR_LOG_MESSAGE
 
 _LOGGER = logging.getLogger(__name__)
 
 MAX_RECONNECTION_ATTEMPTS = 3
+DEVICE_UNAVAILABLE_EXCEPTIONS = (
+    BleakAbortedError,
+    BleakConnectionError,
+    BleakNotFoundError,
+)
 
 
 class SnoozConnectionStatus(Enum):
@@ -35,6 +43,23 @@ class SnoozConnectionStatus(Enum):
     CONNECTED = 2
 
 
+class DisconnectionReason(Enum):
+    # disconnection was initiated by the user
+    USER = 0
+    # the bluetooth connection was lost
+    DEVICE = 1
+    # a connection attempt failed with a known error
+    # like a timeout or device not found
+    DEVICE_UNAVAILABLE = 2
+    # an exception was thrown during connection or
+    # command execution that was uncaught
+    UNEXPECTED_ERROR = 3
+
+
+class SnoozDeviceDisconnectedError(Exception):
+    pass
+
+
 class SnoozDeviceUnavailableError(Exception):
     pass
 
@@ -58,14 +83,16 @@ class SnoozDevice:
         self._device = device
         self._token = token
         self._loop = loop
-        self._connection_ready = Event()
         self._last_dispatched_connection_status = None
+        self._connection_complete = Event()
+        self._connections_exhausted = Event()
         self._connection_attempts: int = 0
         self._connection_start_time: datetime | None = None
         self._connection_ready_time: datetime | None = None
         self._api: SnoozDeviceApi | None = None
         self._connect_lock = Lock()
         self._connection_task: Task[None] | None = None
+        self._reconnection_task: Task[None] | None = None
         self._current_command: SnoozCommandProcessor | None = None
 
         not_disconnected = [
@@ -87,6 +114,7 @@ class SnoozDevice:
             states=states,
             initial=SnoozConnectionStatus.DISCONNECTED,
             after_state_change=self._on_connection_status_change,
+            send_event=True,
         )
         self._machine.add_transition(
             "connection_start",
@@ -108,6 +136,7 @@ class SnoozDevice:
             "device_disconnected",
             not_disconnected,
             SnoozConnectionStatus.DISCONNECTED,
+            after=self._after_device_disconnected,
         )
 
     @property
@@ -118,44 +147,41 @@ class SnoozDevice:
     def address(self) -> str:
         return self._device.address
 
+    @property
+    def connection_status(self) -> SnoozConnectionStatus:
+        return self._machine.connection_status
+
     @property
     def is_connected(self) -> bool:
-        return self._machine.connection_status == SnoozConnectionStatus.CONNECTED
+        return self.connection_status == SnoozConnectionStatus.CONNECTED
 
     async def async_disconnect(self) -> None:
-        if not self.is_connected:
+        if self.connection_status == SnoozConnectionStatus.DISCONNECTED:
             return
 
-        if self._connection_task is not None:
-            self._connection_task.cancel()
-
         self._cancel_current_command()
 
+        if self._reconnection_task is not None and not self._reconnection_task.done():
+            self._reconnection_task.cancel()
+
+        if self._connection_task is not None and not self._connection_task.done():
+            self._connection_task.cancel()
+
         if self._api is not None:
             await self._api.async_disconnect()
 
-        self._machine.device_disconnected()
+        self._machine.device_disconnected(reason=DisconnectionReason.USER)
 
     async def async_execute_command(self, data: SnoozCommandData) -> SnoozCommandResult:
         self._cancel_current_command()
 
         start_time = datetime.now()
-        command = create_command_processor(self._loop, start_time, data)
+        command = create_command_processor(
+            self._loop, start_time, data, format_log_message=self._
+        )
         self._current_command = command
 
-        try:
-            await self._async_wait_for_connection()
-
-            # this shouldn't happen
-            if self._api is None:
-                raise Exception("API is not initialized")
-
-            await command.async_execute(self._api)
-        except SnoozDeviceUnavailableError:
-            command.on_device_unavailable()
-        except Exception:
-            _LOGGER.exception(f"Exception while processing command {command}")
-            command.on_unhandled_exception()
+        await self._async_execute_current_command()
 
         result = await command.result
 
@@ -170,80 +196,90 @@ class SnoozDevice:
         self._current_command.cancel()
         self._current_command = None
 
-    async def _async_wait_for_connection(self) -> None:
-        if self._machine.connection_status == SnoozConnectionStatus.DISCONNECTED:
+    async def _async_execute_current_command(self) -> None:
+        command = self._current_command
+
+        try:
+            await self._async_wait_for_connection_complete()
+
+            if (
+                self._api is not None
+                and command is not None
+                and command.state != CommandProcessorState.COMPLETE
+            ):
+                await command.async_execute(self._api)
+        except CancelledError:
+            # happens when async_disconnect() is called during a connection
+            # we swallow it because we want to escape execution of the command
+            pass
+        except SnoozDeviceDisconnectedError:
+            self._machine.device_disconnected(reason=DisconnectionReason.DEVICE)
+        except SnoozDeviceUnavailableError:
+            self._machine.device_disconnected(
+                reason=DisconnectionReason.DEVICE_UNAVAILABLE
+            )
+        except Exception:
+            self._machine.device_disconnected(
+                reason=DisconnectionReason.UNEXPECTED_ERROR
+            )
+
+    async def _async_wait_for_connection_complete(self) -> None:
+        if self.connection_status == SnoozConnectionStatus.DISCONNECTED:
             async with self._connect_lock:
                 if self._connection_task is None or self._connection_task.done():
                     self._connection_task = self._loop.create_task(
-                        self._async_connect(), name=f"Connect to SNOOZ {self.address}"
+                        self._async_connect(), name=f"[Connect] {self.display_name}"
                     )
             await self._connection_task
 
-        await self._connection_ready.wait()
-
-        # if we reach this state, it means connection attempts are exhausted
-        if self._machine.connection_status != SnoozConnectionStatus.CONNECTED:
-            raise SnoozDeviceUnavailableError()
+        await self._connection_complete.wait()
 
     async def _async_connect(self) -> None:
-        self._machine.connection_start()
+        if self.connection_status == SnoozConnectionStatus.DISCONNECTED:
+            self._machine.connection_start()
 
         try:
-            client = await establish_connection(
-                BleakClient,
-                self._device,
-                self._device.name,
-            )
-        except (BleakConnectionError, BleakNotFoundError, BleakAbortedError):
-            _LOGGER.exception("Device unavailable")
-            raise SnoozDeviceUnavailableError()
+            api = await self._async_create_api()
+            self._api = api
+        except (DEVICE_UNAVAILABLE_EXCEPTIONS) as ex:
+            raise SnoozDeviceUnavailableError() from ex
 
-        api = SnoozDeviceApi(client)
-        api.events.on_disconnect += lambda: self._machine.device_disconnected()
+        api.events.on_disconnect += lambda: self._machine.device_disconnected(
+            reason=DisconnectionReason.DEVICE
+        )
         api.events.on_state_change += lambda state: self._on_receive_device_state(state)
-        self._api = api
-
-        self._machine.device_connected()
 
         await api.async_authenticate_connection(bytes.fromhex(self._token))
-
-        # has to be called after authenticated
         await api.async_listen_for_state_changes()
 
-        # device could disconnect in the previous api call
-        if self._machine.connection_status != SnoozConnectionStatus.CONNECTING:
-            return
+        if self.connection_status == SnoozConnectionStatus.CONNECTING:
+            self._machine.connection_ready()
 
-        self._machine.connection_ready()
+    async def _async_create_api(self) -> SnoozDeviceApi:
+        client = await establish_connection(
+            BleakClient,
+            self._device,
+            self.display_name,
+        )
 
-    async def _async_reconnect(self) -> None:
-        try:
-            await self._async_connect()
+        return SnoozDeviceApi(client)
 
-            # retry executing the current command
-            if self._current_command is not None:
-                # this shouldn't happen
-                if self._api is None:
-                    raise Exception("API is not initialized")
-
-                await self._current_command.async_execute(self._api)
-        except Exception:
-            _LOGGER.exception(f"Exception while reconnecting to {self.address}")
-            self._machine.device_disconnected()
-
-    def _on_connection_start(self) -> None:
+    def _on_connection_start(self, e: EventData) -> None:
+        self._connection_complete.clear()
+        self._connections_exhausted.clear()
         self._connection_attempts += 1
         self._connection_start_time = datetime.now()
 
-    def _on_connection_ready(self) -> None:
+    def _on_connection_ready(self, e: EventData) -> None:
         self._connection_attempts = 0
         self._connection_ready_time = datetime.now()
         self.events.on_connection_load_time(
             datetime.now() - self._connection_ready_time
         )
-        self._connection_ready.set()
+        self._connection_complete.set()
+        self._connections_exhausted.clear()
 
-    def _on_connection_status_change(self) -> None:
+    def _on_connection_status_change(self, e: EventData) -> None:
         new_status = self._machine.connection_status
 
         if new_status == self._last_dispatched_connection_status:
@@ -260,39 +296,86 @@ class SnoozDevice:
         if was_changed:
             self.events.on_state_change(self.state)
 
-    def _on_device_disconnected(self) -> None:
+    def _after_device_disconnected(self, e: EventData) -> None:
+        reason: DisconnectionReason = e.kwargs.get("reason")
+        _LOGGER.debug(
+            self._(f"disconnected because {self._describe(reason)}"),
+            exc_info=reason != DisconnectionReason.USER,
+        )
+
+    def _on_device_disconnected(self, e: EventData) -> None:
+        self._api = None
+
         last_event = self._connection_ready_time or self._connection_start_time
         if last_event is not None:
             self.events.on_connection_duration(datetime.now() - last_event)
 
         self._connection_start_time = None
         self._connection_ready_time = None
-        self._connection_ready.clear()
+        self._connection_complete.set()
 
-        if self._current_command is not None:
-            self._current_command.on_disconnected()
+        disconnect_reason: DisconnectionReason = e.kwargs.get("reason")
+
+        # if the disconnection was initiated from the user, don't attempt to reconnect
+        if disconnect_reason == DisconnectionReason.USER:
+            return
 
         if self._connection_attempts >= MAX_RECONNECTION_ATTEMPTS:
-            # unblock the current command from waiting for a connection
-            if self._current_command is not None:
-                self._connection_ready.set()
-
             _LOGGER.error(
-                f"Device {self.address} unavailable after"
-                f" {self._connection_attempts} attempts"
+                self._(
+                    f"Unavailable after {self._connection_attempts} connection attempts"
+                )
             )
 
+            if (
+                self._current_command is not None
+                and self._current_command.state != CommandProcessorState.COMPLETE
+            ):
+                self._current_command.on_device_unavailable()
+
+            self._connections_exhausted.set()
+            self._connection_attempts = 0
+
             return
 
+        if self._current_command is not None:
+            # don't reconnect on unexpected errors
+            if disconnect_reason == DisconnectionReason.UNEXPECTED_ERROR:
+                self._current_command.on_unhandled_exception()
+                return
+
+            self._current_command.on_disconnected()
+
         # attempt to reconnect automatically
         # we don't await the result since this is called from a sync state transition
         # we cleanup this task on disconnect
-        self._connection_task = self._loop.create_task(
-            self._async_reconnect(), name=f"Reconnect to SNOOZ {self.address}"
+        self._reconnection_task = self._loop.create_task(
+            self._async_execute_current_command(),
+            name=f"[Reconnect] {self.display_name}",
         )
 
+    def _describe(self, reason: DisconnectionReason) -> str:
+        descriptions = {
+            DisconnectionReason.USER: (
+                f"{SnoozDevice.async_disconnect.__qualname__}() was called"
+            ),
+            DisconnectionReason.DEVICE: "the bluetooth connection was lost",
+            DisconnectionReason.UNEXPECTED_ERROR: (
+                f"an uncaught exception occurred.\n{UNEXPECTED_ERROR_LOG_MESSAGE}"
+            ),
+            DisconnectionReason.DEVICE_UNAVAILABLE: (
+                "the device couldn't establish a connection"
+            ),
+        }
+        return descriptions[reason]
+
+    def _(self, message: str) -> str:
+        """Format a message for logging."""
+        return f"[{self.display_name}] ({self.connection_status.name}): {message}"
+
+    @property
     def display_name(self) -> str:
-        return ""
+        return get_snooz_display_name(self.name, self.address)
 
     def __repr__(self) -> str:
         description = []
